import os
import sys
import subprocess
import numpy as np
import torch
import inspect
import json
import copy
import argparse
import random
import wandb

import config
from models.utils import get_model
from data.utils import get_dataset
from optim.base import train_base
from optim.sparse import train_sparse
import distributed
import powersgd

def get_args():
    parser = argparse.ArgumentParser(allow_abbrev=False)
    parser.add_argument('--config_format', default='base', choices=config.registered_formats())

    args, rem_args = parser.parse_known_args()

    return config.parse_args_with_format(format=args.config_format, base_parser=parser, args=rem_args, namespace=args)


def get_exp_name(args):
    """ Returns the name of the experiment, used for saving models and wandb. """
    exp_name = f"{args.model}_lr{args.lr}_bs{args.batch_size}x{args.acc_steps}_{args.world_size}nodes"
    if args.wandb_run_prefix != 'none':
        exp_name = args.wandb_run_prefix + '_' + exp_name
    if 'sparse' in args.model:
        exp_name += f"_lmd{args.lmbda}"
    exp_name += f"_seed={args.seed}"
    return exp_name


def main(args): 

    torch.backends.cuda.matmul.allow_tf32 = True # allows us to make sure we're able to use tensorfloat32 during training
    torch.backends.cudnn.allow_tf32 = True

    distributed_backend = distributed.make_backend_from_args(args)
    args = distributed_backend.get_adjusted_args_for_process(args)

    args.device = torch.device(args.device)
    device_type = "cuda" if "cuda" in str(args.device) else "cpu"
    if device_type == "cuda":
        torch.cuda.set_device(args.device)

    torch.manual_seed(args.seed)
    random.seed(args.seed)
    np.random.seed(args.seed)
    
    print(f"Loading dataset '{args.dataset}'")
    
    data = get_dataset(args) # data is a dict: {'train': train_tokenized, 'val': eval_tokenized}
        
    print(f"Num training tokens: {len(data['train'])}")
    print(f"Num validation tokens: {len(data['val'])}")
    models=[]
    group_specs=[]
    opts=[]
    schedulers=[]
    models.append(get_model(args).to(args.device)) # todo: take care of initializing the model if args.use_pretrained != 'none'
    models[0] = distributed_backend.transform_model(models[0])
    for i in range(1, args.workers):
        m = get_model(args).to(args.device)
        m = distributed_backend.transform_model(m)
        m.load_state_dict(models[0].state_dict())
        models.append(m)
    for i in range(args.workers):
        group_specs.append(distributed_backend.get_raw_model(models[i]).get_parameter_group_specs())
        param_name_mapping = {p_name: p for p_name, p in models[i].named_parameters()}
        optimized_params_cnt = 0
        for g in group_specs[i]:
            params = []
            for p_name in g["params"]:
                translated_p_names = distributed_backend.translate_model_parameter_name_for_node(p_name)
                params += [param_name_mapping[p_name] for p_name in translated_p_names]
            g["params"] = params
            optimized_params_cnt += sum([p.numel() for p in g["params"]])
        print("number of optimized parameters: %.2fM" % (optimized_params_cnt/1e6,))
        if args.opt == 'adamw':
            use_fused = (device_type == 'cuda') and ('fused' in inspect.signature(torch.optim.AdamW).parameters)
            print(f"using fused AdamW: {use_fused}")
            extra_args = dict(fused=True) if use_fused else dict()
            opts.append(torch.optim.AdamW(group_specs[i], lr=args.lr, betas=(args.beta1, args.beta2),
                                    weight_decay=args.weight_decay, **extra_args))
        else:
            opts.append(torch.optim.SGD(group_specs[i], lr=args.lr, momentum=0.9, weight_decay=args.weight_decay))
    
        if args.scheduler != 'none':
            if args.scheduler in ['cos', 'linear']:
                schedulers.append(torch.optim.lr_scheduler.OneCycleLR(optimizer=opts[i], max_lr=args.lr, total_steps=args.iterations, 
                                                                pct_start=args.warmup_percent, anneal_strategy=args.scheduler, 
                                                                cycle_momentum=False, div_factor=1e2, final_div_factor=.05))
            else:
                raise NotImplementedError(f"Unknown scheduler type: {args.scheduler}.")
        else:
            schedulers.append(None)

    for name, p in models[0].named_parameters():
        print(name + str(p.shape))
    
    args.world_size = distributed_backend.get_world_size()
    exp_name = get_exp_name(args)
    if distributed_backend.is_master_process() and args.wandb:
        params_copy = copy.deepcopy(vars(args))
        del params_copy['device']
        wandb.init(project=args.wandb_project, name=exp_name, config=params_copy)
    
    ckpt_path = os.path.join(args.results_base_folder, args.dataset, args.model, exp_name)
    if not os.path.exists(ckpt_path):
        if distributed_backend.is_master_process():
            os.makedirs(ckpt_path)
    elif os.path.isfile(os.path.join(ckpt_path, "summary.json")): # the experiment was already completed
        print(f"Already found experiment '{ckpt_path}'.\nSkipping.")
        sys.exit(0)

    if args.model == 'base': # all train functions have the same interface
        train = train_base
    elif 'sparse' in args.model:
        train = train_sparse
    else:
        raise NotImplementedError(f"No training method implemented for model type '{args.model}'.")

    print(f"\nTraining model={args.model} \n{vars(args)}\n")

    stats = train(models, opts, data, schedulers, args.iterations, args.acc_steps, args.batch_size, args.sequence_length, 
                  eval_freq=args.eval_freq, 
                  distributed_backend=distributed_backend,
                  ckpt_path=f"{ckpt_path}/ckpt.pt", extra_args=args)
    
    args.device = None
    args.dtype = None
    stats['args'] = vars(args)
    if distributed_backend.is_master_process():
        with open(f"{ckpt_path}/summary.json", "w") as fs:
            json.dump(stats, fs)
    distributed_backend.finalize()


if __name__ == "__main__":
    args = get_args()
    main(args)